#include "WinPacketFiltering_PacketFiltering.h"
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <fltdefs.h>

#pragma comment(lib, "iphlpapi.lib")

#define DEBUG 0

void getError(DWORD, LPVOID);
void JBYTEtoPBYTE(jbyte*, PBYTE, int);
void showFilter(PF_FILTER_DESCRIPTOR);
void ByteArrayForkCopy(PBYTE, PBYTE, int);

static INTERFACE_HANDLE iHandle;
const char* sNO_ERROR = "NO_ERROR";

// C^tF[Xnh̍쐬
JNIEXPORT jstring JNICALL Java_WinPacketFiltering_PacketFiltering_PfCreateInterface
	(JNIEnv *env, jclass cls,
		jint		dwName,
		jboolean	inAction_i,
		jboolean	outAction_i,
		jboolean	bUseLog,
		jboolean	bMustBeUnique)
{
	PFFORWARD_ACTION inAction, outAction;
	DWORD result;
	LPVOID lpMsgBuf;
	
	if(inAction_i==JNI_TRUE)	inAction = PF_ACTION_DROP;
	else						inAction = PF_ACTION_FORWARD;
	if(outAction_i==JNI_TRUE)	outAction = PF_ACTION_DROP;
	else						outAction = PF_ACTION_FORWARD;
	
	if(DEBUG) printf("PfCreateInterface : %d, %d, %d, %d, %d, I=%d\n",			// fobO
				dwName, inAction, outAction, bUseLog, bMustBeUnique, &iHandle);	// fobO
	
	result = PfCreateInterface(dwName, inAction, outAction,
							bUseLog, bMustBeUnique, &iHandle);
	
	if(DEBUG) printf("PfCreateInterface : result = %d\n",result);	// fobO
	
	if(result==NO_ERROR)
	{
		return env->NewStringUTF(sNO_ERROR);
	}
	else
	{
		getError(result, &lpMsgBuf);
		return env->NewStringUTF((const char* )lpMsgBuf);
	}
} // PfCreateInterface END

// CoEhpnhɃtB^ǉ
JNIEXPORT jstring JNICALL Java_WinPacketFiltering_PacketFiltering_PfAddFiltersToInterfaceIN
	(JNIEnv *env, jclass cls,
		jint		AddrType,
		jbyteArray	SrcAddr,
		jbyteArray	SrcMask,
		jbyteArray	DstAddr,
		jbyteArray	DstMask,
		jint		SrcPort,
		jint		DstPort,
		jint		CeilingSrcPort,
		jint		CeilingDstPort,
		jint		Protocol)
{
	PF_FILTER_DESCRIPTOR filter;
	DWORD result;
	LPVOID lpMsgBuf;
	PBYTE adrs;
	int	array_size = 0;
	jbyte*	Sadrs;
	
	filter.dwFilterFlags	= 0;
	filter.dwRule			= 0;
	filter.fLateBound		= 0;
	
	if(AddrType==4)
	{
		array_size = 4;
		filter.pfatType	 = PF_IPV4;
		adrs = (PBYTE)malloc(sizeof(BYTE)*array_size);
	}
	else if(AddrType==6)
	{
		array_size = 16;
		filter.pfatType	 = PF_IPV6;
		adrs = (PBYTE)malloc(sizeof(BYTE)*array_size);
	}	
	
	Sadrs = env->GetByteArrayElements(SrcAddr, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.SrcAddr = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.SrcAddr, array_size);
	Sadrs = env->GetByteArrayElements(SrcMask, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.SrcMask = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.SrcMask, array_size);
	Sadrs = env->GetByteArrayElements(DstAddr, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.DstAddr = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.DstAddr, array_size);
	Sadrs = env->GetByteArrayElements(DstMask, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.DstMask = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.DstMask, array_size);
	
	filter.wSrcPort	= SrcPort;
	filter.wDstPort	= DstPort;
	if(CeilingSrcPort == -1) filter.wSrcPortHighRange = SrcPort;
	else					 filter.wSrcPortHighRange = CeilingSrcPort;
	if(CeilingDstPort == -1) filter.wDstPortHighRange = DstPort;
	else					 filter.wDstPortHighRange = CeilingDstPort;
	
	switch(Protocol)
	{
		case 1:	filter.dwProtocol = FILTER_PROTO_ANY; break;
		case 2:	filter.dwProtocol = FILTER_PROTO_ICMP;
				filter.wSrcPort = FILTER_ICMP_TYPE_ANY;
				filter.wDstPort = FILTER_ICMP_TYPE_ANY;
				filter.wSrcPortHighRange = FILTER_ICMP_TYPE_ANY;
				filter.wDstPortHighRange = FILTER_ICMP_TYPE_ANY; break;
		case 3:	filter.dwProtocol = FILTER_PROTO_TCP; break;
		case 4:	filter.dwProtocol = FILTER_PROTO_UDP; break;
	}
	
	if(DEBUG)	// fobO
	{
		printf("PfAddFiltersToInterfaceIN : ");
		showFilter(filter);
		printf("i=%d\n", &iHandle);
	}
	
	result = PfAddFiltersToInterface(iHandle, 1, &filter, 0, NULL, NULL);
	free(adrs);
	free(filter.SrcAddr);
	free(filter.SrcMask);
	free(filter.DstAddr);
	free(filter.DstMask);
	
	if(DEBUG) printf("PfAddFiltersToInterfaceIN : result = %d\n",result);	// fobO
	
	if(result==NO_ERROR)
	{
		return env->NewStringUTF(sNO_ERROR);
	}
	else
	{
		getError(result, &lpMsgBuf);
		return env->NewStringUTF((const char* )lpMsgBuf);
	}
} // PfAddFiltersToInterfaceIN END

// AEgoEhpnhɃtB^ǉ
JNIEXPORT jstring JNICALL Java_WinPacketFiltering_PacketFiltering_PfAddFiltersToInterfaceOUT
	(JNIEnv *env, jclass cls,
		jint		addrType,
		jbyteArray	SrcAddr,
		jbyteArray	SrcMask,
		jbyteArray	DstAddr,
		jbyteArray	DstMask,
		jint		SrcPort,
		jint		DstPort,
		jint		CeilingSrcPort,
		jint		CeilingDstPort,
		jint		Protocol)
{
	PF_FILTER_DESCRIPTOR filter;
	DWORD result;
	LPVOID lpMsgBuf;
	PBYTE adrs;
	int	array_size = 0;
	jbyte*	Sadrs;
	
	filter.dwFilterFlags	= 0;
	filter.dwRule			= 0;
	filter.fLateBound		= 0;
	
	if(addrType==4)
	{
		array_size = 4;
		filter.pfatType	 = PF_IPV4;
		adrs = (PBYTE)malloc(sizeof(BYTE)*array_size);
	}
	else if(addrType==6)
	{
		array_size = 16;
		filter.pfatType	 = PF_IPV6;
		adrs = (PBYTE)malloc(sizeof(BYTE)*array_size);
	}	
	
	Sadrs = env->GetByteArrayElements(SrcAddr, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.SrcAddr = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.SrcAddr, array_size);
	Sadrs = env->GetByteArrayElements(SrcMask, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.SrcMask = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.SrcMask, array_size);
	Sadrs = env->GetByteArrayElements(DstAddr, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.DstAddr = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.DstAddr, array_size);
	Sadrs = env->GetByteArrayElements(DstMask, 0);
	JBYTEtoPBYTE(Sadrs, adrs, array_size);
	filter.DstMask = (PBYTE)malloc(sizeof(BYTE)*array_size);
	ByteArrayForkCopy(adrs, filter.DstMask, array_size);
	
	filter.wSrcPort	= SrcPort;
	filter.wDstPort	= DstPort;
	if(CeilingSrcPort == -1) filter.wSrcPortHighRange = SrcPort;
	else					 filter.wSrcPortHighRange = CeilingSrcPort;
	if(CeilingDstPort == -1) filter.wDstPortHighRange = DstPort;
	else					 filter.wDstPortHighRange = CeilingDstPort;
		
	switch(Protocol)
	{
		case 1:	filter.dwProtocol = FILTER_PROTO_ANY; break;
		case 2:	filter.dwProtocol = FILTER_PROTO_ICMP;
				filter.wSrcPort = FILTER_ICMP_TYPE_ANY;
				filter.wDstPort = FILTER_ICMP_TYPE_ANY;
				filter.wSrcPortHighRange = FILTER_ICMP_TYPE_ANY;
				filter.wDstPortHighRange = FILTER_ICMP_TYPE_ANY; break;
		case 3:	filter.dwProtocol = FILTER_PROTO_TCP; break;
		case 4:	filter.dwProtocol = FILTER_PROTO_UDP; break;
	}
	
	if(DEBUG)	// fobO
	{
		printf("PfAddFiltersToInterfaceOUT : ");
		showFilter(filter);
		printf("i=%d\n", &iHandle);
	}
	
	result = PfAddFiltersToInterface(iHandle, 0, NULL, 1, &filter, NULL);
	free(adrs);
	free(filter.SrcAddr);
	free(filter.SrcMask);
	free(filter.DstAddr);
	free(filter.DstMask);
	
	if(DEBUG) printf("PfAddFiltersToInterfaceOUT : result = %d\n",result);	// fobO
	
	if(result==NO_ERROR)
	{
		return env->NewStringUTF(sNO_ERROR);
	}
	else
	{
		getError(result, &lpMsgBuf);
		return env->NewStringUTF((const char* )lpMsgBuf);
	}
} // PfAddFiltersToInterfaceOUT END

// IPAhXC^tF[XnhɃoCh
JNIEXPORT jstring JNICALL Java_WinPacketFiltering_PacketFiltering_PfBindInterfaceToIPAddress
	(JNIEnv *env, jclass cls,
		jint		AddrType,
		jbyteArray	IPAddress_i)
{
	PBYTE IPAddress;
	int	array_size = 0;
	jbyte*	SIPAddress;
	DWORD result;
	LPVOID lpMsgBuf;
	PFADDRESSTYPE pfatType = NULL;
	
	if(AddrType==4)
	{
		array_size = 4;
		pfatType = PF_IPV4;
		IPAddress = (PBYTE)malloc(sizeof(BYTE)*array_size);
	}
	else if(AddrType==6)
	{
		array_size = 16;
		pfatType = PF_IPV6;
		IPAddress = (PBYTE)malloc(sizeof(BYTE)*array_size);
	}
	
	SIPAddress = env->GetByteArrayElements(IPAddress_i, 0);
	JBYTEtoPBYTE(SIPAddress, IPAddress, array_size);
	
	if(DEBUG)	// fobO
	{
		printf("PfBindInterfaceToIPAddress : %d,  %d.%d.%d.%d  i=%d\n",
		pfatType, IPAddress[0], IPAddress[1],
		IPAddress[2], IPAddress[3], &iHandle);
	}
	
	result = PfBindInterfaceToIPAddress(iHandle, pfatType, IPAddress);
	free(IPAddress);
	
	if(DEBUG) printf("PfBindInterfaceToIPAddress : result = %d\n",result);	// fobO
	
	if(result==NO_ERROR)
	{
		return env->NewStringUTF(sNO_ERROR);
	}
	else
	{
		getError(result, &lpMsgBuf);
		return env->NewStringUTF((const char* )lpMsgBuf);
	}
} // PfBindInterfaceToIPAddress END

// C^tF[Xnh̃AoEh
JNIEXPORT jstring JNICALL Java_WinPacketFiltering_PacketFiltering_PfUnBindInterface
	(JNIEnv *env, jclass cls)
{
	DWORD result;
	LPVOID lpMsgBuf;
	
	result = PfUnBindInterface(iHandle);
	
	if(DEBUG) printf("PfUnBindInterface : result = %d\n",result);
	
	if(result==NO_ERROR)
	{
		return env->NewStringUTF(sNO_ERROR);
	}
	else
	{
		getError(result, &lpMsgBuf);
		return env->NewStringUTF((const char* )lpMsgBuf);
	}
} // PfUnBindInterface END

// C^tF[Xnh̍폜
JNIEXPORT jstring JNICALL Java_WinPacketFiltering_PacketFiltering_PfDeleteInterface
	(JNIEnv *env, jclass cls)
{
	DWORD result;
	LPVOID lpMsgBuf;
	
	result = PfDeleteInterface(iHandle);
	
	if(DEBUG) printf("PfDeleteInterface : result = %d\n",result);	// fobO
	
	if(result==NO_ERROR)
	{
		return env->NewStringUTF(sNO_ERROR);
	}
	else
	{
		getError(result, &lpMsgBuf);
		return env->NewStringUTF((const char* )lpMsgBuf);
	}
} // PfDeleteInterface END

// G[bZ[W̎擾
void getError(DWORD messageID, LPVOID lpMsgBuf)
{
	FormatMessage(
    	FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    	NULL,
    	messageID,
		MAKELANGID(LANG_ENGLISH , SUBLANG_ENGLISH_US),
		(LPTSTR) lpMsgBuf,
		0,
		NULL
	);
} // getError END

// JBYTEzPBYTEzɃRs[
void JBYTEtoPBYTE(jbyte* sB, PBYTE usB, int size)
{
	int i;
	
	for(i=0;i<size;i++)
	{
		if(sB[i]<0)
			usB[i] = sB[i] + 256;
		else
			usB[i] = sB[i];
	}
} // JBYTEtoPBYTE END

// fobOptB^\
void showFilter(PF_FILTER_DESCRIPTOR filter)
{
	int i;
	printf("%d, %d, %d\n",filter.dwFilterFlags, filter.dwRule, filter.pfatType);
	for(i=0;i<4;i++) printf("%d.", filter.SrcAddr[i]);
	printf("	");
	for(i=0;i<4;i++) printf("%d.", filter.SrcMask[i]);
	printf("\n");
	for(i=0;i<4;i++) printf("%d.", filter.DstAddr[i]);
	printf("	");
	for(i=0;i<4;i++) printf("%d.", filter.DstMask[i]);
	printf("\n");
	printf("%d, %d,    %d, %d, %d, %d\n",filter.dwProtocol, filter.fLateBound,
		filter.wSrcPort, filter.wDstPort,
		filter.wSrcPortHighRange, filter.wDstPortHighRange);
} // showFilter END

// PBYTEz̃Rs[
void ByteArrayForkCopy(PBYTE from, PBYTE to, int size)
{
	int i;
	for(i=0;i<size;i++)
		to[i] = from[i];
} // ByteArrayForkCopy END